def get_full_class_name(cls):
    """
    Returns the full name of a class, including the module name.

    :param cls: The class.
    :return: The full name of the class
    """
    return cls.__module__ + "." + cls.__name__


def recursive_subclasses(cls):
    """
    :param cls: The class.
    :return: A list of the classes subclasses.
    """
    return cls.__subclasses__() + [g for s in cls.__subclasses__() for g in recursive_subclasses(s)]



from typing import Any, Iterable, Dict, Tuple

import portion
from typing_extensions import Union

from . import utils

AssignmentType = Union[portion.Interval, Tuple]


class Variable:
    """
    Abstract base class for all variables.
    """

    name: str
    """
    The name of the variable. The name is used for comparison and hashing.
    """

    domain: AssignmentType
    """
    The set of possible events of the variable.
    """

    def __init__(self, name: str, domain: Any):
        self.name = name
        self.domain = domain

    def __lt__(self, other: "Variable") -> bool:
        """
        Returns True if self < other, False otherwise.
        """
        return self.name < other.name

    def __gt__(self, other: "Variable") -> bool:
        """
        Returns True if self > other, False otherwise.
        """
        return self.name > other.name

    def __hash__(self) -> int:
        return self.name.__hash__()

    def __eq__(self, other):
        return self.name == other.name and self.domain == other.domain

    def __str__(self):
        return f"{self.__class__.__name__}({self.name}, {self.domain})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def encode(self, value: Any) -> Any:
        """
        Encode an element of the domain to a representation that is usable for computations.

        :param value: The element to encode
        :return: The encoded element
        """
        return value

    def decode(self, value: Any) -> Any:
        """
        Decode an element to the domain from a representation that is usable for computations.

        :param value: The element to decode
        :return: The decoded element
        """
        return value

    def encode_many(self, elements: Iterable) -> Iterable[Any]:
        """
        Encode many elements of the domain to representations that are usable for computations.

        :param elements: The elements to encode
        :return: The encoded elements
        """
        return elements

    def decode_many(self, elements: Iterable) -> Iterable[Any]:
        """
        Decode many elements from the representations that are usable for computations to their domains.

        :param elements: The encoded elements
        :return: The decoded elements
        """
        return elements

    def to_json(self) -> Dict[str, Any]:
        return {"name": self.name, "type": utils.get_full_class_name(self.__class__), "domain": self.domain}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> 'Variable':
        """
        Create a variable from a json dict.
        This method is called from the from_json method after the correct subclass is determined.

        :param data: The json dict
        :return: The variable
        """
        return cls(name=data["name"], domain=data["domain"])

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> 'Variable':
        """
        Create the correct instanceof the subclass from a json dict.

        :param data: The json dict
        :return: The correct instance of the subclass
        """
        for subclass in utils.recursive_subclasses(Variable):
            if utils.get_full_class_name(subclass) == data["type"]:
                return subclass._from_json(data)

        raise ValueError("Unknown type for variable. Type is {}".format(data["type"]))

    def complement_of_assignment(self, assignment: AssignmentType, encoded: bool = False) -> AssignmentType:
        """
        Returns the complement of the assignment for the variable.

        :param assignment: The assignment
        :param encoded: If the assignment is encoded
        :return: The complement of the assignment
        """
        raise NotImplementedError

    @staticmethod
    def intersection_of_assignments(assignment1: AssignmentType,
                                    assignment2: AssignmentType,
                                    encoded: bool = False) -> AssignmentType:
        """
        Returns the intersection of two assignments

        :param assignment1: The first assignment
        :param assignment2: The second assignment
                :param encoded: If the assignment is encoded
        :return: The intersection of the assignments
        """
        raise NotImplementedError

    @staticmethod
    def union_of_assignments(assignment1: AssignmentType,
                             assignment2: AssignmentType,
                             encoded: bool = False) -> AssignmentType:
            """
            Returns the union of two assignments

            :param assignment1: The first assignment
            :param assignment2: The second assignment
            :param encoded: If the assignment is encoded
            :return: The union of the assignments
            """
            raise NotImplementedError


class Continuous(Variable):
    """
    Class for real valued random variables.
    """

    domain: portion.Interval

    def __init__(self, name: str, domain: portion.Interval = portion.open(-portion.inf, portion.inf)):
        super().__init__(name=name, domain=domain)

    def to_json(self) -> Dict[str, Any]:
        return {"name": self.name, "type": utils.get_full_class_name(self.__class__),
                "domain": portion.to_data(self.domain)}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> 'Variable':
        return cls(name=data["name"], domain=portion.from_data(data["domain"]))

    def complement_of_assignment(self, assignment: portion.Interval, encoded: bool = False) -> portion.Interval:
        return self.domain - assignment

    @staticmethod
    def intersection_of_assignments(assignment1: portion.Interval,
                                    assignment2: portion.Interval,
                                    encoded: bool = False) -> portion.Interval:
        return assignment1 & assignment2

    @staticmethod
    def union_of_assignments(assignment1: portion.Interval,
                             assignment2: portion.Interval,
                             encoded: bool = False) -> portion.Interval:
        return assignment1 | assignment2


class Discrete(Variable):
    """
    Class for discrete countable random variables.
    """
    domain: Tuple

    def __init__(self, name: str, domain: Iterable):
        super().__init__(name=name, domain=tuple(sorted(set(domain))))

    def encode(self, element: Any) -> int:
        """
        Encode an element of the domain to its index.

        :param element: The element to encode
        :return: The index of the element
        """
        return self.domain.index(element)

    def decode(self, index: int) -> Any:
        """
        Decode an index to its element of the domain.

        :param index: The elements index
        :return: The element itself
        """
        return self.domain[index]

    def encode_many(self, elements: Iterable) -> Iterable[int]:
        """
        Encode many elements of the domain to the indices of the elements.

        :param elements: The elements to encode
        :return: The encoded elements
        """
        return tuple(map(self.encode, elements))

    def decode_many(self, elements: Iterable[int]) -> Iterable[Any]:
        """
        Decode many elements from indices to their domains.

        :param elements: The encoded elements
        :return: The decoded elements
        """
        return tuple(map(self.decode, elements))

    def complement_of_assignment(self, assignment: Tuple, encoded: bool = False) -> Tuple:
        if not encoded:
            return tuple(sorted(set(self.domain) - set(assignment)))
        else:
            return tuple(sorted(set(range(len(self.domain))) - set(assignment)))

    @staticmethod
    def intersection_of_assignments(assignment1: Tuple,
                                    assignment2: Tuple,
                                    encoded: bool = False) -> Tuple:

        return tuple(sorted(set(assignment1) & set(assignment2)))

    @staticmethod
    def union_of_assignments(assignment1: Tuple,
                             assignment2: Tuple,
                             encoded: bool = False) -> Tuple:
        return tuple(sorted(set(assignment1) | set(assignment2)))


class Symbolic(Discrete):
    """
    Class for unordered, finite, discrete random variables.
    """
    ...


class Integer(Discrete):
    """Class for ordered, discrete random variables."""
    ...



import unittest

import portion

from random_events.variables import Integer, Symbolic, Continuous, Variable


class VariablesTestCase(unittest.TestCase):
    """Tests for `variables.py`."""

    integer: Integer
    symbol: Symbolic
    real: Continuous

    @classmethod
    def setUpClass(cls):
        """
        Create some variables for testing.
        """
        cls.integer = Integer("integer", set(range(10)))
        cls.symbol = Symbolic("symbol", {"a", "b", "c"})
        cls.real = Continuous("real")

    def test_creation(self):
        """
        Test that the variables are created correctly.
        """
        self.assertEqual(self.integer.name, "integer")
        self.assertEqual(self.integer.domain, tuple(range(10)))
        self.assertEqual(self.symbol.name, "symbol")
        self.assertEqual(self.symbol.domain, ("a", "b", "c"))
        self.assertEqual(self.real.name, "real")
        self.assertEqual(self.real.domain, portion.open(-portion.inf, portion.inf))

    def test_hash(self):
        """
        Test that the variables are hashable.
        """
        self.assertTrue(hash(self.integer))
        self.assertTrue(hash(self.symbol))
        self.assertTrue(hash(self.real))

    def test_ordering(self):
        """
        Test that the variables are ordered correctly.
        """
        self.assertLess(self.integer, self.symbol)
        self.assertLess(self.real, self.symbol)
        self.assertLess(self.integer, self.real)
        self.assertEqual([self.integer, self.real, self.symbol], sorted([self.symbol, self.integer, self.real]))

    def test_equality(self):
        """
        Test that the variables are equal to themselves and not equal to others.
        """
        self.assertEqual(self.integer, Integer("integer", tuple(range(10))))
        self.assertNotEqual(self.symbol, Symbolic("symbol", ("d", "b", "c")))
        self.assertEqual(self.symbol, Symbolic("symbol", ("a", "b", "c")))
        self.assertEqual(self.real, Continuous("real"))

    def test_to_json(self):
        """
        Test that the variables can be dumped to json.
        """
        self.assertTrue(self.symbol.to_json())
        self.assertTrue(self.integer.to_json())
        self.assertTrue(self.real.to_json())

    def test_encode(self):
        """
        Test that the variables can be encoded.
        """
        self.assertEqual(self.integer.encode(1), 1)
        self.assertEqual(self.symbol.encode("b"), 1)
        self.assertEqual(self.real.encode(1.0), 1.0)

    def test_decode(self):
        """
        Test that the variables can be decoded.
        """
        self.assertEqual(self.integer.decode(1), 1)
        self.assertEqual(self.symbol.decode(1), "b")
        self.assertEqual(self.real.decode(1.0), 1.0)

    def test_polymorphic_serialization(self):
        real = Variable.from_json(self.real.to_json())
        self.assertEqual(real, self.real)

        integer = Variable.from_json(self.integer.to_json())
        self.assertEqual(integer, self.integer)

        symbol = Variable.from_json(self.symbol.to_json())
        self.assertEqual(symbol, self.symbol)

    def test_complement_of_assignment(self):
        """
        Test that the complement of an assignment is correct.
        """
        self.assertEqual(self.real.complement_of_assignment(portion.closed(0, 1)),
                         portion.open(-portion.inf, 0) | portion.open(1, portion.inf))
        self.assertEqual(self.symbol.complement_of_assignment(("a",)), ("b", "c", ))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import itertools
from collections import UserDict

import numpy as np
import portion
import plotly.graph_objects as go

from typing_extensions import Set, Union, Any, TYPE_CHECKING, Iterable, List, Self, Dict

from .variables import Variable, Continuous, Discrete


# Type hinting for Python 3.7 to 3.9
if TYPE_CHECKING:
    VariableMapType = UserDict[str, Variable]
else:
    VariableMapType = UserDict


class VariableMap(VariableMapType):
    """
    A map of variables to values.

    Accessing a variable by name is also supported.
    """

    def variable_of(self, name: str) -> Variable:
        """
        Get the variable with the given name.
        :param name: The variable's name
        :return: The variable itself
        """
        variable = [variable for variable in self.keys() if variable.name == name]
        if len(variable) == 0:
            raise KeyError(f"Variable {name} not found in event {self}")
        return variable[0]

    def __getitem__(self, item: Union[str, Variable]):
        if isinstance(item, str):
            item = self.variable_of(item)
        return super().__getitem__(item)

    def __setitem__(self, key: Union[str, Variable], value: Any):
        if isinstance(key, str):
            key = self.variable_of(key)

        if not isinstance(key, Variable):
            raise TypeError(f"Key must be a Variable, not {type(key)}")
        super().__setitem__(key, value)

    def __copy__(self):
        return self.__class__({variable: value for variable, value in self.items()})


# Type hinting for Python 3.7 to 3.9
if TYPE_CHECKING:
    EventMapType = VariableMap[str, Union[tuple, portion.Interval]]
else:
    EventMapType = VariableMap


class SupportsSetOperations:
    """
    A class that supports set operations.
    """

    def union(self, other: Self) -> Self:
        """
        Form the union of this object with another object.
        """
        raise NotImplementedError

    def __or__(self, other: Self):
        return self.union(other)

    def intersection(self, other: Self) -> Self:
        """
        Form the intersection of this object with another object.
        """
        raise NotImplementedError

    def __and__(self, other):
        return self.intersection(other)

    def difference(self, other: Self) -> Self:
        """
        Form the difference of this object with another object.
        """
        raise NotImplementedError

    def __sub__(self, other):
        return self.difference(other)

    def complement(self) -> Self:
        """
        Form the complement of this object.
        """
        raise NotImplementedError

    def __invert__(self):
        return self.complement()

    def is_empty(self) -> bool:
        """
        Check if this object is empty.
        """
        raise NotImplementedError


class Event(SupportsSetOperations, EventMapType):
    """
    A map of variables to values of their respective domains.
    """

    def check_same_type(self, other: Any):
        """
        Check that both self and other are of the same type.

        :param other: The other object
        """
        if type(self) is not type(other):
            raise TypeError(f"Cannot use operation on {type(self)} with {type(other)}")

    def intersection(self, other: EventType) -> EventType:

        # if the other is a complex event
        if isinstance(other, ComplexEvent):

            # flip the call
            return other.intersection(ComplexEvent([self]))

        self.check_same_type(other)
        result = self.__class__()

        variables = set(self.keys()) | set(other.keys())

        for variable in variables:
            assignment1 = self.get(variable, variable.domain)
            assignment2 = other.get(variable, variable.domain)
            intersection = variable.intersection_of_assignments(assignment1, assignment2)
            if len(intersection) == 0:
                return self.__class__()
            result[variable] = intersection

        return result

    def union(self, other: EventType) -> ComplexEvent:
        # create complex event from self
        complex_self = ComplexEvent([self])

        # if the other is a complex event
        if isinstance(other, ComplexEvent):

            # flip the call
            return other.union(complex_self)

        self.check_same_type(other)

        # form the intersection
        intersection = self.intersection(other)

        # if the intersection of the two events is empty
        if intersection.is_empty():

            # trivially mount it
            complex_self.events.append(other)
            return complex_self

        # form complement of intersection
        complement_of_intersection = intersection.complement()

        # intersect the other event with the complement of the intersection
        fragments_of_other = complement_of_intersection.intersection(ComplexEvent([other]))

        # add the fragments of the other event
        complex_self.events.extend(fragments_of_other.events)
        return complex_self

    def difference(self, other: EventType) -> ComplexEvent:
        # if the other is a complex event
        if isinstance(other, ComplexEvent):

            # flip the call
            return other.complement().intersection(ComplexEvent([self]))

        self.check_same_type(other)

        # form the intersection
        intersection = self.intersection(other)

        # if the intersection of the two events is empty
        if intersection.is_empty():
            return ComplexEvent([self])

        # form complement of intersection
        complement_of_intersection = intersection.complement()

        # construct intersection of complement
        return ComplexEvent([event.intersection(self) for event in complement_of_intersection.events
                             if not event.is_empty()])

    def complement(self) -> ComplexEvent:
        # initialize events
        events = []

        # get variables as set
        variables: Set[Variable] = set(self.keys())

        # memorize processed variables
        processed_variables = []

        # for every assignment
        for variable, value in self.items():

            # create the current complementary event
            complement_event = self.__class__()

            # invert this variables assignment
            complement_event[variable] = variable.complement_of_assignment(
                value, encoded=isinstance(self, EncodedEvent))

            # for every other variable
            for other_variable in variables.difference({variable}):

                # if the other variable is already processed
                if other_variable in processed_variables:
                    # add the assignment to the current event
                    complement_event[other_variable] = self[other_variable]
                else:
                    # add the entire domain to the current event
                    other_domain = other_variable.domain

                    # encode if necessary
                    if isinstance(self, EncodedEvent):
                        other_domain = other_variable.encode_many(other_domain)
                    complement_event[other_variable] = other_domain

            # add to processed variables
            processed_variables.append(variable)

            # add to complex event
            if not complement_event.is_empty():
                events.append(complement_event)
        return ComplexEvent(events)

    def __eq__(self, other: Self) -> bool:
        """
        Check if two events are equal.

        If one variable is only in one of the events, it is assumed that the other event has the entire domain as
        default value.
        """

        variables = set(self.keys()) | set(other.keys())

        equal = True

        for variable in variables:
            if variable in self and variable not in other:
                value_equal = variable.domain == self[variable]
            elif variable in other and variable not in self:
                value_equal = variable.domain == other[variable]
            else:
                value_equal = self[variable] == other[variable]
            equal &= value_equal

        return equal

    @staticmethod
    def check_element(variable: Variable, element: Any) -> Union[tuple, portion.Interval]:
        """
        Check that elements can be regarded as elements of the variable's domain.

        Wrap a single element into a set or interval, depending on the variable type.
        For any Iterable type that is not a string, the element is converted to a tuple of that iterable.

        :param variable: The variable where the element should belong to
        :param element: The element to wrap
        :return: The wrapped element
        """

        if isinstance(element, Iterable) and not isinstance(element, (str, portion.Interval)):
            element = tuple(element)

        # if the element is already wrapped
        if isinstance(element, (tuple, portion.Interval)):

            # check that the element is in the variable's domain
            if isinstance(variable, Discrete):
                if not all(elem in variable.domain for elem in element):
                    # raise an error
                    raise ValueError(f"Element {element} not in domain {variable.domain}")

                element = tuple(sorted(element))

            # return the element directly
            return element

        # if the element is not in the variables' domain
        if element not in variable.domain:
            # raise an error
            raise ValueError(f"Element {element} not in domain {variable.domain}")

        # if the variable is continuous
        if isinstance(variable, Continuous):
            # return the element as a singleton interval
            return portion.singleton(element)

        # if the variable is discrete
        elif isinstance(variable, Discrete):
            # return the element as a set
            return (element,)
        else:
            raise TypeError(f"Unknown variable type {type(variable)}")

    def __setitem__(self, key: Union[str, Variable], value: Any):
        EventMapType.__setitem__(self, key, self.check_element(key, value))

    def encode(self) -> 'EncodedEvent':
        """
        Encode the event to an encoded event.
        :return: The encoded event
        """
        return EncodedEvent({variable: variable.encode_many(element) for variable, element in self.items()})

    def is_empty(self) -> bool:
        return any(len(value) == 0 for value in self.values()) or len(self.keys()) == 0

    def plot(self) -> Union[List[go.Scatter], List[go.Mesh3d]]:
        """
        Plot the event.
        """
        assert all(isinstance(variable, Continuous) for variable in self.keys()), "Can only plot continuous events"
        if len(self.keys()) == 2:
            return self.plot_2d()
        elif len(self.keys()) == 3:
            return self.plot_3d()
        else:
            raise ValueError("Can only plot 2D and 3D events")

    def plot_2d(self) -> List[go.Scatter]:
        """
        Plot the event in 2D.
        """

        # form cartesian product of all intervals
        intervals = [value._intervals for value in self.values()]
        simple_events = list(itertools.product(*intervals))
        traces = []

        # for every atomic interval
        for simple_event in simple_events:

            # plot a rectangle
            points = np.asarray(list(itertools.product(*[[axis.lower, axis.upper] for axis in simple_event])))
            y_points = points[:, 1]
            y_points[len(y_points) // 2:] = y_points[len(y_points) // 2:][::-1]
            traces.append(go.Scatter(x=np.append(points[:, 0], points[0, 0]), y=np.append(y_points, y_points[0]),
                                     mode="lines", name="Event", fill="toself"))
        return traces

    def plot_3d(self) -> List[go.Mesh3d]:
        """
        Plot the event in 3D.
        """

        # form cartesian product of all intervals
        intervals = [value._intervals for value in self.values()]
        simple_events = list(itertools.product(*intervals))
        traces = []

        # shortcut for the dimensions
        x, y, z = 0, 1, 2

        # for every atomic interval
        for simple_event in simple_events:

            # Create a 3D mesh trace for the rectangle
            traces.append(go.Mesh3d(
                # 8 vertices of a cube
                x=[simple_event[x].lower, simple_event[x].lower, simple_event[x].upper, simple_event[x].upper,
                   simple_event[x].lower, simple_event[x].lower, simple_event[x].upper, simple_event[x].upper],
                y=[simple_event[y].lower, simple_event[y].upper, simple_event[y].upper, simple_event[y].lower,
                   simple_event[y].lower, simple_event[y].upper, simple_event[y].upper, simple_event[y].lower],
                z=[simple_event[z].lower, simple_event[z].lower, simple_event[z].lower, simple_event[z].lower,
                   simple_event[z].upper, simple_event[z].upper, simple_event[z].upper, simple_event[z].upper],
                # i, j and k give the vertices of triangles
                i=[7, 0, 0, 0, 4, 4, 6, 6, 4, 0, 3, 2],
                j=[3, 4, 1, 2, 5, 6, 5, 2, 0, 1, 6, 3],
                k=[0, 7, 2, 3, 6, 7, 1, 1, 5, 5, 7, 6],
                flatshading=True
            ))
        return traces

    def plotly_layout(self) -> Dict:
        """
        Create a layout for the plotly plot.
        """
        variables = list(self.keys())
        if len(variables) == 2:
            result = {"xaxis_title": variables[0].name,
                      "yaxis_title": variables[1].name}
        elif len(variables) == 3:
            result = dict(scene=dict(
                xaxis_title=variables[0].name,
                yaxis_title=variables[1].name,
                zaxis_title=variables[2].name)
            )
        else:
            raise NotImplementedError("Can only plot 2D and 3D events")

        return result

    def __hash__(self):
        return hash(tuple(sorted(self.items())))


class EncodedEvent(Event):
    """
    A map of variables to indices of their respective domains.
    """

    @staticmethod
    def check_element(variable: Variable, element: Any) -> Union[tuple, portion.Interval]:

        # if the variable is continuous
        if isinstance(variable, Continuous):

            # if it's not an interval
            if not isinstance(element, portion.Interval):

                # try to convert it to one
                element = portion.singleton(element)

            return element

        # if its any kind of iterable that's not an interval convert it to a tuple
        if isinstance(element, Iterable) and not isinstance(element, portion.Interval):
            element = tuple(sorted(element))

        # if it is just an int, convert it to a tuple containing the int
        elif isinstance(element, int):
            element = (element, )

        if not isinstance(element, tuple):
            raise ValueError("Element for a discrete domain must be a tuple, not {}".format(type(element)))

        # if any element is not in the index set of the domain, raise an error
        if not all(0 <= elem < len(variable.domain) for elem in element):
            raise ValueError(f"Element {element} not in the index set of the domain {variable.domain}")

        return element

    def decode(self) -> Event:
        """
        Decode the event to a normal event.
        :return: The decoded event
        """
        return Event({variable: variable.decode_many(value) for variable, value in self.items()})


class ComplexEvent(SupportsSetOperations):
    """
    A complex event is a set of mutually exclusive events.
    """
    events: List[Event]

    def __init__(self, events: Iterable[Event]):
        self.events = list(event for event in events if not event.is_empty())

    def union(self, other: Self) -> Self:
        result = ComplexEvent(self.events + other.events)
        return result.make_events_disjoint().simplify()

    def make_events_disjoint(self) -> Self:
        """
        Make all events in this complex event disjoint.

        This is done by forming the intersection of all events recursively until no more intersections are found.
        Then, the original events are decomposed into their disjoint components.
        Finally, a complex event is formed from the disjoint components. Note that the result may not be the minimal
        representation of the complex event, so it is advised to call ``simplify`` afterward.
        """

        # initialize previous intersections
        previous_intersections = []

        # for every pair of events
        for index, event in enumerate(self.events):
            for other_event in self.events[index + 1:]:

                # append intersection of pairwise events
                intersection = event.intersection(other_event)
                if not intersection.is_empty() and intersection not in previous_intersections:
                    previous_intersections.append(intersection)

        # if there are no intersections, skip the rest
        if len(previous_intersections) == 0:
            return self

        # while
        while len(previous_intersections) > 0:

            # initialize new intersections
            new_intersections = []

            # form pairwise intersections of previous intersections
            for index, intersection in enumerate(previous_intersections):
                for other_intersection in previous_intersections[index + 1:]:
                    if not intersection.intersection(other_intersection).is_empty():
                        new_intersections.append(intersection)

            if len(new_intersections) == 0:
                break

            previous_intersections = new_intersections

        # sanity check
        complex_event_of_intersections = ComplexEvent(previous_intersections)
        assert complex_event_of_intersections.are_events_disjoint(), "Events are not disjoint"

        # initialize result
        result = ComplexEvent(complex_event_of_intersections.events)

        # for every original event
        for original_event in self.events:

            # initialize the difference of the original events with the intersection
            decomposed_original_events = set()

            # for every atomic intersection in the disjoint intersections
            for atomic_intersection in complex_event_of_intersections.events:

                # get the difference of the original event with the atomic intersection
                original_event_disjoint_component = original_event.difference(atomic_intersection)

                # unify the differences
                decomposed_original_events = decomposed_original_events.union(
                    set(original_event_disjoint_component.events))

            # add the differences to the result
            result.events.extend(decomposed_original_events)
        return result

    def simplify(self) -> Self:
        """
        Simplify the complex event such that sub-unions of events that can be expressed as a single events
        are merged.
        """

        # for every pair of events
        for index, event in enumerate(self.events):
            for other_event in self.events[index + 1:]:

                # for every variable in the event
                for variable, value in event.items():

                    # if the events match in this dimension
                    if other_event[variable] == value:

                        # form the simpler union of the two events
                        unified_event = Event({variable: value})
                        for variable_ in event.keys():
                            if variable_ != variable:
                                unified_event[variable_] = variable.union_of_assignments(event[variable_],
                                                                                         other_event[variable_])
                        # recurse into the simpler complex event
                        result = ComplexEvent([])
                        result.events.append(unified_event)
                        result.events.extend([event__ for event__ in self.events if event__ != event
                                              and event__ != other_event])
                        return result.simplify()

        # if no simplification is possible, return the current complex event
        return self.__copy__()

    def intersection(self, other: Self) -> Self:
        intersections = [event.intersection(other_event) for other_event in other.events for event in self.events]
        return ComplexEvent(intersections)

    def difference(self, other: Self) -> Self:
        return self.intersection(other.complement())

    def complement(self) -> Self:
        result = self.events[0].complement()
        for event in self.events[1:]:
            current_complement = event.complement()
            result = result.intersection(current_complement)
        return result.make_events_disjoint().simplify()

    def are_events_disjoint(self) -> bool:
        """
        Check if all events inside this complex event are disjoint.
        """
        for index, event in enumerate(self.events):
            for event_ in self.events[index + 1:]:
                if not event.intersection(event_).is_empty():
                    return False
        return True

    def __str__(self):
        return " u ".join(str(event) for event in self.events)

    def __repr__(self):
        return f"Union of {len(self.events)} events"

    def __eq__(self, other: ComplexEvent) -> bool:
        """
        Check if two complex events are equal.
        """
        return (all(event in other.events for event in self.events)
                and all(event in self.events for event in other.events))

    def __copy__(self):
        return self.__class__([event.copy() for event in self.events])

    def plot(self) -> Union[List[go.Scatter], List[go.Mesh3d]]:
        """
        Plot the complex event.
        """
        traces = []
        for event in self.events:
            traces.extend(event.plot())
        return traces

    def plotly_layout(self) -> Dict:
        """
        Create a layout for the plotly plot.
        """
        return self.events[0].plotly_layout()

    def is_empty(self) -> bool:
        return len(self.events) == 0


EventType = Union[Event, EncodedEvent, ComplexEvent]


import unittest

import portion

from random_events.events import VariableMap, Event, EncodedEvent, ComplexEvent
from random_events.variables import Continuous, Integer, Symbolic

import plotly.graph_objects as go


class VariableTestCase(unittest.TestCase):

    integer: Integer
    symbol: Symbolic
    real: Continuous
    event: VariableMap

    @classmethod
    def setUpClass(cls):
        """
        Create some event for testing.
        """
        cls.integer = Integer("integer", set(range(10)))
        cls.symbol = Symbolic("symbol", {"a", "b", "c"})
        cls.real = Continuous("real")
        cls.event = VariableMap({cls.integer: 1, cls.symbol: "a", cls.real: 1.0})

    def test_creation(self):
        """
        Test that the event is created correctly.
        """
        self.assertEqual(self.event[self.integer], 1)
        self.assertEqual(self.event[self.symbol], "a")
        self.assertEqual(self.event[self.real], 1.0)

    def test_string_access(self):
        """
        Test that the event is accessible by string.
        """
        self.event["integer"] = self.event[self.integer]
        self.assertEqual(self.event["integer"], self.event[self.integer])
        self.assertEqual(self.event["symbol"], self.event[self.symbol])
        self.assertEqual(self.event["real"], self.event[self.real])

    def test_raising(self):
        """
        Test that the event raises an error if the variable is not in the map.
        """
        self.assertRaises(KeyError, lambda: self.event["not_in_map"])


class EventTestCase(unittest.TestCase):

    integer: Integer
    symbol: Symbolic
    real: Continuous
    event: Event

    @classmethod
    def setUpClass(cls):
        """
        Create some event for testing.
        """
        cls.integer = Integer("integer", set(range(10)))
        cls.symbol = Symbolic("symbol", {"a", "b", "c"})
        cls.real = Continuous("real")
        cls.event = Event({cls.integer: 1, cls.symbol: "a", cls.real: 1.0})

    def test_wrapping(self):
        """
        Test that the event is wrapped correctly.
        """
        self.assertEqual(self.event[self.integer], (1,))
        self.assertEqual(self.event[self.symbol], ("a",))
        self.assertEqual(self.event[self.real], portion.singleton(1.0))

    def test_set_assignment(self):
        """
        Test that the event is set correctly.
        """
        event = self.event.copy()
        event[self.integer] = (2, 3)
        self.assertEqual(event[self.integer], (2, 3))
        event[self.symbol] = ("b", "c")
        self.assertEqual(event[self.symbol], ("b", "c"))
        event[self.real] = portion.closed(0.0, 1.0)
        self.assertEqual(event[self.real], portion.closed(0.0, 1.0))

    def test_raising(self):
        """
        Test that errors are raised correctly.
        """
        event = self.event.copy()
        with self.assertRaises(ValueError):
            event[self.integer] = 11

        with self.assertRaises(ValueError):
            event[self.integer] = (-1,)

        with self.assertRaises(ValueError):
            event[self.symbol] = "d"

        with self.assertRaises(ValueError):
            event[self.symbol] = ("d",)

    def test_encode(self):
        """
        Test that events are correctly encoded.
        """
        encoded = self.event.encode()
        self.assertIsInstance(encoded, EncodedEvent)
        decoded = encoded.decode()
        self.assertEqual(self.event, decoded)

    def test_intersection(self):
        """
        Test ordinary intersection of events
        """
        event_1 = Event()
        event_1[self.integer] = (1, 2)
        event_1[self.symbol] = ("a", "b")

        result = event_1.intersection(self.event)

        self.assertEqual(result["integer"], (1, ))
        self.assertEqual(result["symbol"], ("a", ))
        self.assertEqual(result["real"], self.event["real"])

    def test_empty_intersection(self):
        """
        Test empty intersection of events
        """
        event_1 = Event()
        event_1[self.integer] = (1, 2)
        event_1[self.symbol] = ("c", )
        result = event_1.intersection(self.event)
        self.assertTrue(result.is_empty())

    def test_intersection_alias(self):
        event_1 = Event()
        event_1[self.integer] = (1, 2)
        event_1[self.symbol] = ("a", "b")
        self.assertEqual(event_1 & self.event, event_1.intersection(self.event))
        self.assertEqual(event_1 & self.event, self.event & event_1)

    def test_union_without_intersection(self):
        event1 = Event({self.integer: 1, self.symbol: "a", self.real: 2.0})
        union = event1.union(self.event)
        self.assertIsInstance(union, ComplexEvent)
        self.assertEqual(len(union.events), 2)
        self.assertTrue(union.are_events_disjoint())

    def test_difference(self):
        event_1 = Event()
        event_1[self.integer] = (1, 2, 5)
        event_1[self.symbol] = ("a", "b")
        result = event_1.difference(self.event)
        self.assertEqual(len(result.events), 3)
        self.assertTrue(result.are_events_disjoint())

    def test_difference_alias(self):
        event_1 = Event()
        event_1[self.integer] = (1, 2, 5)
        event_1[self.symbol] = ("a", "b")
        self.assertEqual(event_1 - self.event, event_1.difference(self.event))
        # differences are not symmetric
        self.assertNotEqual(event_1 - self.event, self.event - event_1)

    def test_equality(self):
        self.assertEqual(self.event, self.event)
        self.assertNotEqual(self.event, Event())

    def test_raises_on_operation_with_different_types(self):
        with self.assertRaises(TypeError):
            self.event & self.event.encode()

        with self.assertRaises(TypeError):
            self.event | self.event.encode()

        with self.assertRaises(TypeError):
            self.event - self.event.encode()


class EncodedEventTestCase(unittest.TestCase):

    integer: Integer
    symbol: Symbolic
    real: Continuous

    @classmethod
    def setUpClass(cls):
        """
        Create some event for testing.
        """
        cls.integer = Integer("integer", set(range(10)))
        cls.symbol = Symbolic("symbol", {"a", "b", "c"})
        cls.real = Continuous("real")

    def test_creation(self):
        event = EncodedEvent()
        event[self.integer] = 1
        self.assertEqual(event[self.integer], (1,))
        event[self.integer] = (1, 2)
        self.assertEqual(event[self.integer], (1, 2))
        event[self.symbol] = 0
        self.assertEqual(event[self.symbol], (0, ))
        event[self.symbol] = {1, 0}
        self.assertEqual(event[self.symbol], (0, 1))

        interval = portion.open(0, 1)
        event[self.real] = interval
        self.assertEqual(interval, event[self.real])

    def test_raises(self):
        event = EncodedEvent()
        with self.assertRaises(ValueError):
            event[self.symbol] = 3

        with self.assertRaises(ValueError):
            event[self.symbol] = portion.open(0, 1)

        with self.assertRaises(ValueError):
            event[self.symbol] = (1, 2, 3, 4)

    def test_dict_like_creation(self):
        event = EncodedEvent(zip([self.integer, self.symbol], [1, 0]))
        self.assertEqual(event[self.integer], (1,))
        self.assertEqual(event[self.symbol], (0,))

        event = EncodedEvent(zip([self.integer, self.symbol], [[0, 1], 0]))
        self.assertEqual(event[self.integer], (0, 1))
        self.assertEqual(event[self.symbol], (0,))

    def test_set_operations_return_type(self):
        event = EncodedEvent(zip([self.integer, self.symbol], [1, 0]))
        self.assertEqual(type(event & event), EncodedEvent)
        self.assertEqual(type(event | event), ComplexEvent)
        self.assertEqual(type(event - event), ComplexEvent)


class ComplexEventTestCase(unittest.TestCase):

    x: Continuous = Continuous("x")
    y: Continuous = Continuous("y")
    z: Continuous = Continuous("z")
    a: Symbolic = Symbolic("a", {"a1", "a2"})
    b: Symbolic = Symbolic("b", {"b1", "b2", "b3"})
    unit_interval = portion.closed(0, 1)

    def test_union(self):
        event_1 = Event({self.x: self.unit_interval, self.y: self.unit_interval})
        event_2 = Event({self.x: portion.closed(0.5, 2), self.y: portion.closed(0.5, 2)})
        union = event_1.union(event_2)
        self.assertIsInstance(union, ComplexEvent)
        self.assertEqual(len(union.events), 3)
        self.assertTrue(union.are_events_disjoint())
        # go.Figure(union.plot()).show()

    def test_union_of_complex_events(self):
        event_1 = Event({self.x: self.unit_interval, self.y: self.unit_interval})
        event_2 = Event({self.x: portion.closed(0.5, 2), self.y: portion.closed(0.5, 2)})
        complex_event_1 = event_1.union(event_2)
        event_3 = Event({self.x: portion.closed(0.5, 2), self.y: portion.closed(-0.5, 2)})
        complex_event_2 = event_1.union(event_3)

        result = complex_event_1.union(complex_event_2)

        self.assertIsInstance(result, ComplexEvent)
        self.assertTrue(result.are_events_disjoint())

    def test_make_events_disjoint_and_simplify(self):
        event_1 = Event({self.x: self.unit_interval, self.y: self.unit_interval})
        event_2 = Event({self.x: portion.closed(0.5, 2), self.y: portion.closed(0.5, 2)})
        complex_event = ComplexEvent([event_1, event_2])
        self.assertFalse(complex_event.are_events_disjoint())
        complex_event = complex_event.make_events_disjoint()
        self.assertEqual(len(complex_event.events), 5)
        self.assertTrue(complex_event.are_events_disjoint())
        simplified_event = complex_event.simplify()
        self.assertEqual(len(simplified_event.events), 3)
        self.assertTrue(simplified_event.are_events_disjoint())

    def test_are_events_disjoint(self):
        event1 = Event({self.x: portion.closed(0, 1), self.y: portion.closed(0, 1)})
        event2 = Event({self.x: portion.closed(0.5, 2), self.y: portion.closed(0.5, 2)})
        event3 = Event({self.x: portion.closed(2, 3), self.y: portion.closed(2, 3)})

        complex_event = ComplexEvent((event1, event2))
        self.assertFalse(complex_event.are_events_disjoint())

        complex_event = ComplexEvent((event1, event3))
        self.assertTrue(complex_event.are_events_disjoint())

        complex_event = ComplexEvent((event1, event3, event2,))
        self.assertFalse(complex_event.are_events_disjoint())

    def test_from_continuous_complement(self):
        event = Event({self.x: portion.closed(0, 1), self.y: portion.closed(0, 1)})
        complement = event.complement()

        self.assertEqual(len(complement.events), 2)
        self.assertTrue(complement.are_events_disjoint())

        c1 = complement.events[0]
        self.assertEqual(c1[self.x], portion.open(-portion.inf, 0) | portion.open(1, portion.inf))
        self.assertEqual(c1[self.y], portion.open(-portion.inf, portion.inf))

        c2 = complement.events[1]
        self.assertEqual(c2[self.y], portion.open(-portion.inf, 0) | portion.open(1, portion.inf))
        self.assertEqual(c2[self.x], event[self.x])

        for sub_event in complement.events:
            self.assertTrue(sub_event.intersection(event).is_empty())

    def test_complement_3d(self):
        event = Event({self.x: portion.closed(0, 1),
                       self.y: portion.closed(0, 1),
                       self.z: portion.closed(0, 1)})
        complement = event.complement()
        self.assertEqual(len(complement.events), 3)
        self.assertTrue(complement.are_events_disjoint())

    def test_from_discrete_complement(self):
        event = Event({self.a: "a1", self.b: "b1"})
        complement = event.complement()

        self.assertEqual(len(complement.events), 2)
        self.assertTrue(complement.are_events_disjoint())

        c1 = complement.events[0]
        self.assertEqual(c1[self.a], ("a2", ))
        self.assertEqual(c1[self.b], self.b.domain)

        c2 = complement.events[1]
        self.assertEqual(c2[self.a], ("a1", ))
        self.assertEqual(c2[self.b], ("b2", "b3"))

        for sub_event in complement.events:
            self.assertTrue(sub_event.intersection(event).is_empty())

    def test_chained_complement(self):
        event = Event({self.x: portion.closed(0, 1), self.y: portion.closed(0, 1)})
        complement = event.complement()
        copied_event = complement.complement()
        self.assertEqual(len(copied_event.events), 1)
        self.assertEqual(copied_event.events[0], event)


class PlottingTestCase(unittest.TestCase):
    x: Continuous = Continuous("x")
    y: Continuous = Continuous("y")
    z: Continuous = Continuous("z")

    def test_plot_2d(self):
        event = Event({self.x: portion.closed(0, 1), self.y: portion.closed(0, 1)})
        fig = go.Figure(event.plot())
        # fig.show()

    def test_plot_3d(self):
        event = Event({self.x: portion.closed(0, 1), self.y: portion.closed(0, 1), self.z: portion.closed(0, 1)})
        fig = go.Figure(event.plot())
        # fig.show()

    def test_plot_complex_event_2d(self):
        event = Event({self.x: portion.closed(0, 1), self.y: portion.closed(0, 1)})
        complement = event.complement()
        limiting_event = Event({self.x: portion.closed(-1, 2), self.y: portion.closed(-1, 2)})
        result = complement.intersection(ComplexEvent([limiting_event]))
        fig = go.Figure(result.plot(), result.plotly_layout())
        # fig.show()

    def test_plot_complex_event_3d(self):
        event = Event({self.x: portion.closed(0, 1),
                       self.y: portion.closed(0, 1),
                       self.z: portion.closed(0, 1)})
        complement = event.complement()
        limiting_event = Event({self.x: portion.closed(-1, 2),
                                self.y: portion.closed(-1, 2),
                                self.z: portion.closed(-1, 2)})
        result = complement.intersection(ComplexEvent([limiting_event]))
        fig = go.Figure(result.plot(), result.plotly_layout())
        # fig.show()


if __name__ == '__main__':
    unittest.main()
